import numpy as np
from kshape.core import _sbd
from cobras_ts.superinstance import SuperInstance


def get_prototype(A,indices, prototype):
    max_dist_to_others = np.inf
    prototype_idx = None

    for idx in indices:
        cur_dist, _ = _sbd(A[idx,:],prototype)
        if cur_dist < max_dist_to_others: 
            max_dist_to_others = cur_dist
            prototype_idx = idx
    return prototype_idx

class SuperInstance_kShape(SuperInstance):

    def __init__(self, data, indices, train_indices, sbd_centroid=None):
        super(SuperInstance_kShape, self).__init__(data, indices, train_indices)

        self.sbd_centroid = sbd_centroid

        if sbd_centroid is not None:
            self.representative_idx = get_prototype(data, self.train_indices, sbd_centroid)
        else:
            self.representative_idx = indices[0]

    def distance_to(self, other_cluster):
        d, _ = _sbd(self.sbd_centroid, other_cluster.sbd_centroid)
        return d



import argparse
import sys
from pathlib import Path
import logging
import subprocess
import time
import platform

import numpy as np
from sklearn import metrics


logger = logging.getLogger("cobra_ts")


description = """COBRAS-TS Time Series Active Semi-Supervised Clustering."""
epilog = """
Copyright 2018 KU Leuven, DTAI Research Group.
"""


def prepare_data(inputs, fileformat, labelcol, **_kwargs):
    """Read dataset based on the arguments given by the user.

    :param inputs: List of filepaths
    :param fileformat: Type of datafile (choices are "csv")
    :param labelcol: Integer or none
    :return: (series, labels)
    """
    data_fn = Path(inputs[0])
    data_format = None
    if fileformat is None:
        if data_fn.suffix == '.csv':
            data_format = 'csv'
    else:
        data_format = fileformat

    if data_format == 'csv':
        data = np.loadtxt(str(data_fn), delimiter=',')
    else:
        raise Exception("Unknown file format (use the --format argument)")

    if labelcol is None:
        series = data
        labels = None
    else:
        nonlabelcols = list(idx for idx in range(data.shape[1]) if idx != labelcol)
        series = data[:, nonlabelcols]
        labels = data[:, labelcol]
    return series, labels


def prepare_clusterer(dist, series, querier, budget, dtw_window=None, dtw_alpha=None, **_kwargs):
    """Create a clusterer based on the arguments given by the user.

    :param dist: Type of distance (options: "dtw", "kshape")
    :param series: List of sequences
    :param querier: Querier object
    :param budget: Max number of queries (passed to querier object)
    :param dtw_window: window parameter for DTW  (passed to querier object)
    :param dtw_alpha: Alpha paramter for DTW  (passed to querier object)
    :return: Clusterer object
    """
    if dist == 'dtw':
        from dtaidistance import dtw
        from cobras_ts.cobras_dtw import COBRAS_DTW
        window = dtw_window
        alpha = dtw_alpha
        dists = dtw.distance_matrix(series, window=int(0.01 * window * series.shape[1]))
        dists[dists == np.inf] = 0
        dists = dists + dists.T - np.diag(np.diag(dists))
        # noinspection PyUnresolvedReferences
        affinities = np.exp(-dists * alpha)
        clusterer = COBRAS_DTW(affinities, querier, budget)
    elif dist == 'kshape':
        from cobras_ts.cobras_kshape import COBRAS_kShape
        clusterer = COBRAS_kShape(series, querier, budget)
    else:
        raise Exception("Unknown distance type: {}".format(dist))
    return clusterer


def main(argv=None):
    parser = argparse.ArgumentParser(description=description,
                                     epilog=epilog)

    parser.add_argument('--verbose', '-v', action='count', default=0, help='Verbose output')

    dist_group = parser.add_argument_group("distance arguments")
    dist_group.add_argument('--dist', choices=['dtw', 'kshape'], default='kshape',
                            help='Distance computation')
    dist_group.add_argument('--dtw-window', metavar='INT', dest='dtw_window', type=int, default=10,
                            help='Window size for DTW')
    dist_group.add_argument('--dtw-alpha', metavar='FLOAT', dest='dtw_alpha', type=float, default=0.5,
                            help='Compute affinity from distance: affinity = exp(-dist * alpha)')

    data_group = parser.add_argument_group("dataset arguments")
    data_group.add_argument('--format', dest='fileformat', choices=['csv'], help='Dataset format')
    data_group.add_argument('--labelcol', metavar='INT', type=int,
                            help='Column with labels')
    data_group.add_argument('--visual', action='store_true',
                            help='Use visual interface to query constraints if no labels are given')

    parser.add_argument('--budget', type=int, default=100,
                        help='Number of constraints to ask maximally')
    parser.add_argument('inputs', nargs=1, help='Dataset file')
    args = parser.parse_args(argv)

    logger.setLevel(max(logging.WARNING - 10 * args.verbose, logging.DEBUG))
    logger.addHandler(logging.StreamHandler(sys.stdout))

    if args.visual:
        # TODO: check with Toon, this was in labelcol != none?
        # if platform.system() == "Darwin":
        #     # It's a Mac
        #     logger.info("Opening http://localhost:5006/webapp")
        #     subprocess.check_call(["open", "http://localhost:5006/webapp"])
        webapp_dir = Path(__file__).parent / "webapp"
        logger.debug(f"Opening bokeh webapp at {webapp_dir}")
        # TODO: All arguments should be passed to webapp
        subprocess.check_call(["bokeh", "serve", "--show", str(webapp_dir), "--args", args.inputs[0]])
        logger.info("Bokeh server closed")
        sys.exit(1)

    series, labels = prepare_data(**vars(args))

    if args.labelcol is None:
        from cobras_ts.commandlinequerier import CommandLineQuerier
        querier = CommandLineQuerier()
    else:
        from cobras_ts.labelquerier import LabelQuerier
        querier = LabelQuerier(labels)

    clusterer = prepare_clusterer(series=series, querier=querier, **vars(args))

    logger.info("Start clustering ...")
    start_time = time.time()
    clusterings, runtimes, ml, cl = clusterer.cluster()
    end_time = time.time()
    logger.info("... done clustering in {} seconds".format(end_time - start_time))
    print("Clustering:")
    print(clusterings)
    if args.labelcol is not None:
        print("ARI score = " + str(metrics.adjusted_rand_score(clusterings[-1], labels)))



class Clustering:

    def __init__(self,clusters):
        self.clusters = clusters

    def construct_cluster_labeling(self):

        pts_per_cluster = [cluster.get_all_points() for cluster in self.clusters]

        pred = [-1] * sum([len(x) for x in pts_per_cluster])

        for i, pts in enumerate(pts_per_cluster):
            for pt in pts:
                pred[pt] = i

        return pred





import numpy as np
from kshape.core import kshape, _sbd
from cobras_ts.superinstance_kshape import SuperInstance_kShape

from cobras_ts.cobras import COBRAS


class COBRAS_kShape(COBRAS):

    def split_superinstance(self, si, k):
        # the actual splitting
        pred = kshape(self.data[si.indices, :], k)

        # making sure that all of the super-instances contain at least one training instance
        # super-instances without training instance are merged with the closest one that does contain a
        # training instance
        training = []
        no_training = []

        for new_si_centroid, new_si_idx in pred:
            # go from super instance indices to global ones
            cur_indices = [si.indices[idx] for idx in new_si_idx]

            si_train_indices = [x for x in cur_indices if x in self.train_indices]
            if len(si_train_indices) != 0:
                training.append(SuperInstance_kShape(self.data, cur_indices, self.train_indices, new_si_centroid))
            else:
                no_training.append((cur_indices, new_si_centroid))

        for indices, centroid in no_training:
            # sets of indices without a training point are merged with their closest super-instance
            # closeness is based on the SBD centroid
            closest_train = None
            closest_train_dist = np.inf
            for training_si in training:
                cur_dist, _ = _sbd(training_si.sbd_centroid, centroid)
                if cur_dist < closest_train_dist:
                    closest_train_dist = cur_dist
                    closest_train = training_si
            closest_train.indices.extend(indices)

        return training

    def create_superinstance(self, indices):
        return SuperInstance_kShape(self.data, indices, self.train_indices)


from cobras_ts.querier import Querier

class LabelQuerier(Querier):

    def __init__(self, labels):
        super(LabelQuerier, self).__init__()
        self.labels = labels

    def query_points(self, idx1, idx2):
        return self.labels[idx1] == self.labels[idx2]

    def update_clustering(self, clustering):
        return



import os

import numpy as np
from dtaidistance import dtw
from sklearn import metrics

from cobras_ts.cobras_dtw import COBRAS_DTW
from cobras_ts.labelquerier import LabelQuerier

ucr_path = '/home/toon/Downloads/UCR_TS_Archive_2015'
dataset = 'ECG200'
budget = 100
alpha = 0.5
window = 10

data = np.loadtxt(os.path.join(ucr_path,dataset,dataset + '_TEST'), delimiter=',')
series = data[:,1:]
labels = data[:,0]


dists = dtw.distance_matrix(series, window=int(0.01 * window * series.shape[1]))
dists[dists == np.inf] = 0
dists = dists + dists.T - np.diag(np.diag(dists))
affinities = np.exp(-dists * alpha)

clusterer = COBRAS_DTW(affinities, LabelQuerier(labels), budget)
clusterings, runtimes, ml, cl = clusterer.cluster()
print(clusterings)
print("done")
print(metrics.adjusted_rand_score(clusterings[-1],labels))


"""
Most of the clustering procedure is shared between COBRAS_dtw and COBRAS_kshape, this is captured in the COBRAS class.
The following methods are specific to each variant:

- create_superinstance: super-instances for COBRAS_dtw and COBRAS_kshape are different, this method
                        simply creates a super-instance of the appropriate type
- split_superinstance: this is also different or COBRAS_dtw and COBRAS_kshape, the first uses
                        spectral clustering, the second kshape to split a super-instance
"""

import abc
import itertools
import random
import time

import numpy as np
from cobras_ts.cluster import Cluster

from cobras_ts.clustering import Clustering

class COBRAS:
    def __init__(self, data, querier, max_questions, train_indices=None):
        """COBRAS clustering

        :param data: Data set numpy array of size (nb_instances,nb_features)
        :param querier: Querier object that answers whether instances are linked through a must-link or cannot-link.
        :param max_questions: Maximum number of questions that are asked. Clustering stops after this.
        :param train_indices:
            Default: None
        """
        self.data = data
        self.querier = querier
        self.max_questions = max_questions

        if train_indices is None:
            self.train_indices = range(self.data.shape[0])
        else:
            self.train_indices = train_indices

        self.clustering = None
        self.split_cache = dict()
        self.start = None
        self.results = None
        self.ml = None
        self.cl = None

    def cluster(self):
        """Perform clustering.

        :return: 4-tuple with:
            * List of lists, each list represents the cluster labels in each iteration
            * List of elapsed times, one per iteration
            * List of used must-link constraints
            * List of used cannot-link constraints
        """
        # TODO: Why is a cluster at each iteration kept. Expensive?
        self.start = time.time()

        # 'results' will contain tuples (cluster labels, elapsed time, number of pairwise constraints)
        # we will add an entry for each constraint that is queried
        self.results = [([0] * self.data.shape[0],0,0)]
        self.ml = []
        self.cl = []

        # initially, there is only one super-instance that contains all data indices (i.e. list(range(self.data.shape[0])))
        initial_superinstance = self.create_superinstance(list(range(self.data.shape[0])))

        # the split level for this initial super-instance is determined,
        # the super-instance is split, and a new cluster is created for each of the newly created superinstances
        initial_k = self.determine_split_level(initial_superinstance)

        superinstances = self.split_superinstance(initial_superinstance,initial_k)
        self.clustering = Clustering([])
        for si in superinstances:
            self.clustering.clusters.append(Cluster([si]))

        # in results we store a tuple for each constraint number
        # for the first couple of queries (those used to determine the initial splitting level)
        # we do not have a clustering, we just return 'all elements in one cluster'
        for i in range(len(self.ml) + len(self.cl)):
            self.results.append(([0] * self.data.shape[0], time.time() - self.start, len(self.ml) + len(self.cl)))

        # the first bottom up merging step
        self.merge_containing_clusters(starting_level=True)

        while len(self.ml) + len(self.cl) < self.max_questions:

            self.querier.update_clustering(self.clustering)

            to_split, originating_cluster = self.identify_superinstance_to_split()
            if to_split is None:
                break


            originating_cluster.super_instances.remove(to_split)
            if len(originating_cluster.super_instances) == 0:
                self.clustering.clusters.remove(originating_cluster)

            split_level = self.determine_split_level(to_split)
            new_super_instances = self.split_superinstance(to_split, split_level)
            new_clusters = self.add_new_clusters_from_split(new_super_instances)



            if not new_clusters:
                # it is possible that splitting a super-instance does not lead to a new cluster:
                # e.g. a super-instance constains 2 points, of which one is in the test set
                # in this case, the super-instance can be split into two new ones, but these will be joined
                # again immediately, as we cannot have super-instances containing only test points (these cannot be
                # queried)
                # this case handles this, we simply add the super-instance back to its originating cluster,
                # and set the already_tried flag to make sure we do not keep trying to split this superinstance
                originating_cluster.super_instances.append(to_split)
                to_split.tried_splitting = True
                continue
            else:
                self.clustering.clusters.extend(new_clusters)

            self.merge_containing_clusters(starting_level=False)


        return [clust for clust, _, _ in self.results], [runtime for _, runtime, _ in self.results], self.ml, self.cl

    @abc.abstractmethod
    def split_superinstance(self, si, k):
        return

    @abc.abstractmethod
    def create_superinstance(self, indices):
        return

    def determine_split_level(self, superinstance):
        # need to make a 'deep copy' here, we will split this one a few times just to determine an appropriate splitting
        # level
        si = self.create_superinstance(superinstance.indices)

        must_link_found = False

        split_level = 0
        while not must_link_found:

            if len(si.indices) == 2:
                new_si = [self.create_superinstance([si.indices[0]]), self.create_superinstance([si.indices[1]])]
            else:
                new_si = self.split_superinstance(si,2)

            new_clusters = []
            for si in new_si:
                new_clusters.append(Cluster([si]))

            if len(new_clusters) == 1:
                # we cannot split any further along this branch, we reached the splitting level
                split_level = max([split_level, 1])
                split_n = 2 ** int(split_level)
                return min(len(si.indices),split_n)

            x = new_clusters[0]
            y = new_clusters[1]
            bc1, bc2 = x.get_comparison_points(y)
            pt1 = min([bc1.representative_idx, bc2.representative_idx])
            pt2 = max([bc1.representative_idx, bc2.representative_idx])

            if self.querier.query_points(pt1,pt2):
                self.ml.append((pt1, pt2))
                self.results.append((self.results[-1][0], time.time() - self.start, len(self.ml) + len(self.cl)))
                must_link_found = True
            else:
                self.cl.append((pt1, pt2))
                self.results.append((self.results[-1][0], time.time() - self.start, len(self.ml) + len(self.cl)))
                split_level += 1

            si_to_choose = []
            if len(x.super_instances[0].train_indices) >= 2:
                si_to_choose.append(x.super_instances[0])
            if len(y.super_instances[0].train_indices) >= 2:
                si_to_choose.append(y.super_instances[0])

            if len(si_to_choose) == 0:
                split_level = max([split_level, 1])
                split_n = 2 ** int(split_level)
                return min(len(si.indices), split_n)

            si = random.choice(si_to_choose)

        split_level = max([split_level, 1])
        split_n = 2 ** int(split_level)
        return min(len(si.indices), split_n)

    def add_new_clusters_from_split(self, si):
        new_clusters = []
        for x in si:
            new_clusters.append(Cluster([x]))

        if len(new_clusters) == 1:
            return None
        else:
            return new_clusters

    def merge_containing_clusters(self, starting_level=False):

        start_clustering = self.results[-1][0]

        merged = True
        while merged and len(self.ml) + len(self.cl) < self.max_questions:

            clusters_to_consider = [cluster for cluster in self.clustering.clusters if not cluster.is_finished]

            cluster_pairs = itertools.combinations(clusters_to_consider, 2)
            cluster_pairs = [x for x in cluster_pairs if
                             not x[0].cannot_link_to_other_cluster(x[1], self.cl)]
            cluster_pairs = sorted(cluster_pairs, key=lambda x: x[0].distance_to(x[1]))

            merged = False
            for x, y in cluster_pairs:

                if x.cannot_link_to_other_cluster(y, self.cl):
                    continue

                bc1, bc2 = x.get_comparison_points(y)
                pt1 = min([bc1.representative_idx, bc2.representative_idx])
                pt2 = max([bc1.representative_idx, bc2.representative_idx])


                if (pt1, pt2) in self.ml:
                    x.super_instances.extend(y.super_instances)
                    self.clustering.clusters.remove(y)
                    merged = True
                    break

                if len(self.ml) + len(self.cl) == self.max_questions:
                    break

                if self.querier.query_points(pt1,pt2):
                    x.super_instances.extend(y.super_instances)
                    self.clustering.clusters.remove(y)
                    self.ml.append((pt1, pt2))
                    merged = True

                    if starting_level:
                        # if it is the first merging step there is no previous clustering that is being refined,
                        # so temporary results are the ones being constructed now
                        self.results.append(
                            (self.clustering.construct_cluster_labeling(), time.time() - self.start,
                             len(self.ml) + len(self.cl)))
                    else:
                        self.results.append((start_clustering, time.time() - self.start, len(self.ml) + len(self.cl)))

                    break
                else:
                    self.cl.append((pt1, pt2))

                    if starting_level:
                        self.results.append(
                            (self.clustering.construct_cluster_labeling(), time.time() - self.start,
                             len(self.ml) + len(self.cl)))
                    else:
                        self.results.append((start_clustering, time.time() - self.start, len(self.ml) + len(self.cl)))


            if not starting_level:
                self.results[-1] = (self.clustering.construct_cluster_labeling(), time.time() - self.start,
                                    len(self.ml) + len(self.cl))


    def identify_superinstance_to_split(self):

        if len(self.clustering.clusters) == 1 and len(self.clustering.clusters[0].super_instances) == 1:
            return self.clustering.clusters[0].super_instances[0], self.clustering.clusters[0]

        superinstance_to_split = None
        max_heur = -np.inf
        originating_cluster = None

        for cluster in self.clustering.clusters:

            if cluster.is_pure:
                continue

            if cluster.is_finished:
                continue

            for superinstance in cluster.super_instances:
                if superinstance.tried_splitting:
                    continue

                if len(superinstance.indices) > max_heur:
                    superinstance_to_split = superinstance
                    max_heur = len(superinstance.indices)
                    originating_cluster = cluster

        if superinstance_to_split is None:
            return None, None


        return superinstance_to_split, originating_cluster





import numpy as np
from sklearn.cluster import SpectralClustering
from cobras_ts.superinstance_dtw import SuperInstance_DTW, get_prototype

from cobras_ts.cobras import COBRAS


class COBRAS_DTW(COBRAS):

    def split_superinstance(self, si, k):
        data_to_cluster = self.data[np.ix_(si.indices, si.indices)]
        spec = SpectralClustering(k, affinity="precomputed")
        spec.fit(data_to_cluster)
        split_labels = spec.labels_.astype(np.int)

        labels_to_indices = []
        for label in set(split_labels):
            labels_to_indices.append(np.where(split_labels == label))

        training = []
        no_training = []

        for new_si_idx in range(len(set(split_labels))):
            # go from super instance indices to global ones
            cur_indices = [si.indices[idx] for idx, c in enumerate(split_labels) if c == new_si_idx]

            si_train_indices = [x for x in cur_indices if x in self.train_indices]
            if len(si_train_indices) != 0:
                training.append(SuperInstance_DTW(self.data, cur_indices, self.train_indices))
            else:
                no_training.append((cur_indices, get_prototype(self.data, cur_indices)))

        for indices, centroid in no_training:
            closest_train = max(training, key=lambda x: self.data[x.representative_idx, centroid])
            closest_train.indices.extend(indices)

        return training

    def create_superinstance(self, indices):
        return SuperInstance_DTW(self.data, indices, self.train_indices)


import os

import numpy as np
from sklearn import metrics

from cobras_ts.cobras_kshape import COBRAS_kShape
from cobras_ts.labelquerier import LabelQuerier

ucr_path = '/home/toon/Downloads/UCR_TS_Archive_2015'
dataset = 'ECG200'
budget = 100

data = np.loadtxt(os.path.join(ucr_path,dataset,dataset + '_TEST'), delimiter=',')
series = data[:,1:]
labels = data[:,0]

clusterer = COBRAS_kShape(series, LabelQuerier(labels), budget)
clusterings, runtimes, ml, cl = clusterer.cluster()

print(clusterings)
print("done")
print(metrics.adjusted_rand_score(clusterings[-1],labels))


import os, sys
from threading import Thread
from bokeh.layouts import column, row
from bokeh.models import Button
from bokeh.plotting import curdoc, figure
from bokeh.models.widgets import Div

try:
    import datashader
except ImportError:
    datashader = None
    print("\n\nThe datashader package needs to be installed from source to use the GUI:\n"
          "$ pip install git+ssh://git@github.com/bokeh/datashader.git@0.6.5#egg=datashader-0.6.5\n\n")
if datashader is None:
    sys.exit(1)

try:
    from cobras_ts.visualquerier import VisualQuerier
    from cobras_ts.cobras_kshape import COBRAS_kShape
except ImportError:
    sys.path.append(os.path.join(os.path.abspath(os.path.dirname(__file__)), os.pardir, os.pardir))
    from cobras_ts.visualquerier import VisualQuerier
    from cobras_ts.cobras_kshape import COBRAS_kShape

import random
import numpy as np
import pandas as pd
import sys




curdoc().title = "COBRAS-TS"

loading = Div(text="""<h3>Loading...<h3>""", width=100, height=100)

def mustlink_callback():
    global query_answered
    global querier
    global layout
    querier.query_answered = True
    querier.query_result = True
    layout.children[1].children[1].children[1] = loading

def cannotlink_callback():
    global query_answered
    global querier
    global layout
    querier.query_answered = True
    querier.query_result = False
    layout.children[1].children[1].children[1] = loading

button_ml = Button(label="Yes (must-link)", button_type="success")
button_ml.on_click(mustlink_callback)

button_cl = Button(label="No (cannot-link)", button_type="warning")
button_cl.on_click(cannotlink_callback)


random.seed(123)
np.random.seed(123)

query_answered = False

fn = sys.argv[1]
doc = curdoc()

# TODO: We should reuse cli.prepare_data() here, it is now hardcoded for one case
df = pd.read_csv(fn,header=None)

labels = df.ix[:,0]
df = df.drop(0,axis=1)
ts = df.ix[0,:]

data = df.as_matrix()

# reformat the data into an appropriate DataFrame
dfs = []
split = pd.DataFrame({'x': [np.nan]})
for i in range(len(data)):
    x = list(range(len(ts)))
    y = data[i]
    df3 = pd.DataFrame({'x': x, 'y': y})
    dfs.append(df3)
    dfs.append(split)
df2 = pd.concat(dfs, ignore_index=True)

x_range = 0, data.shape[1]
y_range = data[1:].min(), data[1:].max()

all_data_plot = figure(plot_width=400, plot_height=180, x_range=x_range, y_range=y_range, title="Full dataset",toolbar_location='above')
p = figure(plot_width=400, plot_height=180, x_range=x_range, y_range=y_range, title="Full dataset",toolbar_location='above')
canvas = datashader.Canvas(x_range=x_range, y_range=y_range,
                           plot_height=180, plot_width=400)
agg = canvas.line(df2, 'x', 'y', datashader.count())
img = datashader.transfer_functions.shade(agg, how='eq_hist')
all_data_plot.image_rgba(image=[img.data], x=x_range[0], y=y_range[0], dw=x_range[1] - x_range[0], dh=y_range[1] - y_range[0])
p.image_rgba(image=[img.data], x=x_range[0], y=y_range[0], dw=x_range[1] - x_range[0], dh=y_range[1] - y_range[0])
initial_temp_clustering = row(p)


topdiv = Div(text="<h1> COBRAS<sup>TS</sup> <br>  iteration:  1 <br> # queries answered: 0 </h1>", css_classes=['top_title_div'],
width=500, height=100)
div = Div(text="<h2> The full dataset </h2>", css_classes=['title_div'],
width=200, height=100)
div2 = Div(text="<h2> Should these two instances be in the same cluster? </h2>", css_classes=['title_div'],
width=500, height=100)
div3 = Div(text="<h2> The (intermediate) clustering </h2>", css_classes=['title_div'],
width=400, height=100)
div4 = Div(text="", css_classes=['title_div'],width=400, height=100)


ts1 = figure(x_axis_type="datetime", plot_width=250, plot_height=120, toolbar_location=None) # placeholders
ts2 = figure(x_axis_type="datetime", plot_width=250, plot_height=120, toolbar_location=None)

layout = column(row(topdiv), row(column(div, all_data_plot), column(div2, row(ts1, ts2), column(button_ml, button_cl))), div3, initial_temp_clustering, div4)
curdoc().add_root(layout)


querier = VisualQuerier(data, curdoc(), layout)
clusterer = COBRAS_kShape(data, querier, 100000)

def blocking_task():
    clusterer.cluster()

thread = Thread(target=blocking_task)
thread.start()





import abc


class SuperInstance:

    def __init__(self, data, indices, train_indices):
        if not isinstance(indices, list):
            raise ValueError('Should give a list of indices as input to SuperInstance...')

        self.data = data
        self.indices = indices
        self.train_indices = [x for x in indices if x in train_indices]
        self.tried_splitting = False

        self.representative_idx = None

    def get_representative_idx(self):
        try:
            return self.representative_idx
        except:
            raise ValueError('Super instances without training instances')

    @abc.abstractmethod
    def distance_to(self, other_cluster):
        return


from cobras_ts.querier import Querier

from tornado import gen
from bokeh.plotting import figure
from bokeh.layouts import row, column
from bokeh.models import Button, Toggle
from bokeh.models.widgets import Div

from functools import partial
import time
import numpy as np
import pandas as pd
import datashader


colors = ["#cc6600", "#a0a0a0", "#00cccc", "#0066cc", "#0000cc"]


@gen.coroutine
def update(bokeh_layout, xs, ys1, ys2, iteration, num_queries):

    topdiv = Div(text="<font size=\"15\"> <b>COBRAS<sup>TS</sup></b> </font>  <br><font size=\"2\">  # queries answered: " + str(num_queries) + "</font>", css_classes=['top_title_div'],
        width=500, height=100)
    bokeh_layout.children[0].children[0] = column(topdiv)

    # This is the only thing that does not give glitches: making new figures each time.
    # Much better would probably be to put a figure once, and simply update the plot data.
    # This does not work, however, lines seem to frequently disappear for some unknown reason.
    ts1 = figure(x_axis_type="datetime", plot_width=250, plot_height=120, toolbar_location=None)
    ts2 = figure(x_axis_type="datetime", plot_width=250, plot_height=120, toolbar_location=None)
    ts1.line('x', 'y', source=dict(x=xs, y=ys1), line_width=1)
    ts2.line('x', 'y', source=dict(x=xs, y=ys2), line_width=1)

    bokeh_layout.children[1].children[1].children[1] = row(ts1,ts2)


def cluster_is_pure(metadata, attr, old_value, new_value):
    metadata["cluster"].is_pure = not metadata["cluster"].is_pure


def cluster_is_finished(metadata, attr, old_value, new_value):
    metadata["cluster"].is_finished = not metadata["cluster"].is_finished


@gen.coroutine
def remove_cluster_indicators(querier, bokeh_layout):
    bokeh_layout.children[4] = row()
    for col in bokeh_layout.children[3].children:
        col.children[1] = row()
        col.children[2] = row()
    querier.finished_indicating = True


def remove_cluster_indicators_callback(querier, bokeh_layout,bokeh_doc):
    bokeh_doc.add_next_tick_callback(partial(remove_cluster_indicators, querier=querier, bokeh_layout=bokeh_layout))


@gen.coroutine
def update_clustering(querier, bokeh_layout, bokeh_doc, data, clustering, cluster_indices, representatives):

    plot_width = int(800 / len(clustering))
    plot_height = int(plot_width / 2)

    plots = []
    cols = []

    for c, c_idxs, cluster_representatives in zip(clustering, cluster_indices, representatives):

        cur_data = data[c_idxs, :]

        x_range = 0, cur_data.shape[1]
        y_range = cur_data[1:].min(), cur_data[1:].max()

        cluster_plot = figure(plot_width=plot_width, plot_height=plot_height, x_range=x_range, y_range=y_range,
                              toolbar_location=None)
        cluster_plot.toolbar.logo = None

        # actually make the plot
        canvas = datashader.Canvas(x_range=x_range, y_range=y_range,
                                   plot_height=plot_height, plot_width=plot_width)

        # reformat the data into an appropriate DataFrame
        dfs = []
        split = pd.DataFrame({'x': [np.nan]})
        # for i in range(len(data)-1):
        for i in range(len(cur_data)):
            x = list(range(cur_data.shape[1]))
            y = cur_data[i]
            df3 = pd.DataFrame({'x': x, 'y': y})
            dfs.append(df3)
            dfs.append(split)
        df2 = pd.concat(dfs, ignore_index=True)

        agg = canvas.line(df2, 'x', 'y', datashader.count())
        img = datashader.transfer_functions.shade(agg, how='eq_hist')

        cluster_plot.image_rgba(image=[img.data], x=x_range[0], y=y_range[0], dw=x_range[1] - x_range[0],
                                dh=y_range[1] - y_range[0])

        for i, repr_idx in enumerate(cluster_representatives):
            cluster_plot.line('x', 'y', source=dict(x=range(data.shape[1]), y=data[repr_idx, :]), line_width=2, line_color=colors[i % len(colors)])

        plots.append(cluster_plot)

        button = Toggle(label="This cluster is pure.", active=c.is_pure)
        button.on_change("active", partial(cluster_is_pure, {"cluster" : c}))

        button2 = Toggle(label="This cluster is pure and complete.", active=c.is_finished)
        button2.on_change("active", partial(cluster_is_finished, {"cluster" : c}))

        cols.append(column(cluster_plot,button,button2))

    topdiv = Div(
        text="<font size=\"15\"> <b>COBRAS<sup>TS</sup></b> </font>  <br><font size=\"2\">  # queries answered: " + str(
            querier.n_queries) + "</font>", css_classes=['top_title_div'],
        width=500, height=100)
    bokeh_layout.children[0].children[0] = column(topdiv)

    bokeh_layout.children[3] = row(cols)

    finished_indicating = Button(label="Show me more queries!",button_type="danger")
    finished_indicating.on_click(partial(remove_cluster_indicators_callback, querier=querier, bokeh_layout=bokeh_layout, bokeh_doc=bokeh_doc))

    bokeh_layout.children[4] = row(finished_indicating)

    bokeh_layout.children[1].children[1].children[1] = Div(text="""Click 'Show me more queries!' below to continue querying""", width=400, height=100)


class VisualQuerier(Querier):

    def __init__(self, data, bokeh_doc, bokeh_layout):
        super(VisualQuerier, self).__init__()

        self.data = data

        self.bokeh_doc = bokeh_doc
        self.bokeh_layout = bokeh_layout

        self.query_answered = False
        self.query_result = None

        self.iteration = 0
        self.n_queries = 0

        self.finished_indicating = False

    def query_points(self, idx1, idx2):

        time.sleep(0.8)  # to fix (?) mysterious issue with bokeh..
        self.bokeh_doc.add_next_tick_callback(
            partial(update, bokeh_layout=self.bokeh_layout, xs=list(range(self.data.shape[1])), ys1=self.data[idx1, :], ys2=self.data[idx2, :], iteration=self.iteration, num_queries=self.n_queries))

        while not self.query_answered:
            pass
        self.query_answered = False

        self.n_queries += 1

        return self.query_result

    def finished_indicating(self):
        return self.finished_indicating

    def update_clustering(self, clustering):

        self.finished_indicating = False

        # we basically have to cache everything here, as it all can be modified in the main cobras loop while
        # the plotting code is running
        clusters = []
        cluster_indices = []
        si_representatives = []
        for cluster in clustering.clusters:
            clusters.append(cluster)
            cluster_indices.append(cluster.get_all_points())
            si_representatives.append([si.representative_idx for si in cluster.super_instances])

        time.sleep(0.8)  # to fix (?) mysterious issue with bokeh..
        self.bokeh_doc.add_next_tick_callback(
            partial(update_clustering, querier=self, bokeh_layout=self.bokeh_layout, bokeh_doc=self.bokeh_doc, data=self.data, clustering=clusters, cluster_indices=cluster_indices, representatives=si_representatives))

        while not self.finished_indicating:
            # this is to prevent cobras from continuing
            pass

        self.iteration += 1




import itertools


class Cluster:

    def __init__(self, super_instances):
        self.super_instances = super_instances
        self.is_pure = False # in the visual querier, the user can indicate that the entire cluster is pure
        self.is_finished = False

    def distance_to(self, other_cluster):
        super_instance_pairs = itertools.product(self.super_instances, other_cluster.super_instances)
        return min([x[0].distance_to(x[1]) for x in super_instance_pairs])

    def get_comparison_points(self, other_cluster):
        # any super-instance should do, no need to find closest ones!
        return self.super_instances[0], other_cluster.super_instances[0]

    def get_all_points(self):
        all_pts = []
        for super_instance in self.super_instances:
            all_pts.extend(super_instance.indices)
        return all_pts

    def cannot_link_to_other_cluster(self, c, cl):
        medoids_c1 = [si.representative_idx for si in self.super_instances]
        medoids_c2 = [si.representative_idx for si in c.super_instances]

        for x, y in itertools.product(medoids_c1, medoids_c2):
            if (x, y) in cl or (y, x) in cl:
                return True
        return False


import numpy as np

from cobras_ts.superinstance import SuperInstance


def get_prototype(A,indices):
    max_affinity_to_others = -np.inf
    prototype_idx = None

    for idx in indices:
        affinity_to_others = 0.0
        for j in indices:
            if j == idx:
                continue
            affinity_to_others += A[idx,j]
        if affinity_to_others > max_affinity_to_others:
            prototype_idx = idx
            max_affinity_to_others = affinity_to_others

    return prototype_idx


class SuperInstance_DTW(SuperInstance):

    def __init__(self, data, indices, train_indices):
        super(SuperInstance_DTW, self).__init__(data, indices, train_indices)
        self.representative_idx = get_prototype(self.data, self.train_indices)

    def distance_to(self, other_cluster):
        # not really a distance, but that does not matter for COBRAS execution, this is only used for sorting etc.
        return -self.data[self.representative_idx, other_cluster.representative_idx]



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='cobras_ts',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.5',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Semi-supervised time series clustering with COBRAS',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://bitbucket.org/toon_vc/cobras_ts/',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Toon Van Craenendonck',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='toon.vancraenendonck@cs.kuleuven.be',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='clustering timeseries semi-supervised pairwise constraints',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['dtaidistance', 'kshape', 'numpy', 'scikit-learn[alldeps]'],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'dev': ['check-manifest', 'sphinx_rtd_theme'],
        'test': ['coverage'],
        'gui': ['bokeh', 'cloudpickle', 'datashader']  # datashader has no pipy entry
        # include --find-links https://people.cs.kuleuven.be/wannes.meert/pip/pip.html
        # url dependencies are note allowed anymore in pip 10
        # 'pip @ git+ssh://git@github.com/bokeh/datashader.git@0.6.5#egg=datashader-0.6.5']
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={  # Optional
        'sample': ['package_data.dat'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        'console_scripts': [
            'cobras_ts=cobras_ts.cli:main',
        ],
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional

    },
)



import abc


class Querier:

    def __init__(self):
        pass

    @abc.abstractmethod
    def query_points(self, idx1, idx2):
        return

    @abc.abstractmethod
    def update_clustering(self, clustering):
        # not ideal? this has not too much to do with querying, it is only needed for the webapp
        return

    @abc.abstractmethod
    def update_clustering_detailed(self, clustering):
        # not ideal? this has not too much to do with querying, it is only needed for the webapp
        return
